module ebox_io

  use param  

  use dal
  use sla
  use eboxdata
  use edetect_io_box
  use edetect_gen
  !!! IT 2015/06 for detToRaDec
  use cal
  use cal_aux       ! requires caloalutils in DEPEND and cal in src/Makefile
  use coordinate    ! requires ssclib in DEPEND and ssc in src/Makefile

  implicit none
  private

  public wr_stab, boxout, rd_data, dmpari
 
!  character(len=16), dimension(max_ima) ,public :: instrume 
  character(len=21), dimension(max_ima) ,public :: instrume
  real(kind=double), dimension(max_ima), public :: energy 
  character(1024), dimension(max_ima) :: c_image, c_bgima, c_expima, c_detima
  real(kind=single), dimension(max_ima), public :: &
                                         xcenters, ycenters, exp_center 

contains

  !---------------------------------------------------------------------
  subroutine dmpari 
  !---------------------------------------------------------------------

    character(len=1024), dimension(max_ima) :: imag
    real :: likmin
    integer :: num_runs, i, necf, npi
    integer, dimension(max_ima) :: band
    logical :: locflg, matchflg, maskflg, expoflg
    !!--JR
    logical :: withimbuff
 
    type(datasett) :: set
    type(tablet) :: tab
    
    locflg = .not. booleanParameter("usemap")
    matchflg = booleanParameter("usematchedfilter")
    maskflg = booleanParameter("withdetmask")
    expoflg = booleanParameter("withexpimage")
    !!--- JR -----
    withimbuff = booleanParameter("withimagebuffersize")
    !!------------
    num_runs = intParameter("nruns")
    likmin = realParameter("likemin")
    do i=1, parameterCount("hrdef")
      hrdef(i) = intParameter("hrdef",i-1)
    end do
    boxsize = intParameter("boxsize")

    if (boxsize < 4) then
      boxsize = 1
    else 
      boxsize = 2
    end if

    !!-- JR ------
    if (withimbuff) then
      imbuff = intParameter("imagebuffersize")
      withimbuff = .true.
      else
      imbuff = 640
      withimbuff = .false.
    endif
    !!------------

    !!! IT 2015/03 new parameter obsmode: pointing or slew-mode data
    !!! (def.: pointing; alt.: slew; no other values allowed by config file)
    obsmode = stringParameter("obsmode")

    xoff=0.
    yoff=0.
    if (booleanParameter("withoffsets")) then
      set = dataSet(stringParameter("mergedListset"), READ)
      if (hasAttribute(set,"XOFFPNM1")) then    
        xoff(2)=Real32Attribute(set,"XOFFPNM1")
      else 
        tab = table(set,"SRCLIST")
        if (hasAttribute(tab,"XOFFPNM1")) then
          xoff(2)=Real32Attribute(tab,"XOFFPNM1")
        end if        
      end if
      if (hasAttribute(set,"XOFFPNM2")) then    
        xoff(3)=Real32Attribute(set,"XOFFPNM2")
      else 
        tab = table(set,"SRCLIST")
        if (hasAttribute(tab,"XOFFPNM2")) then
          xoff(3)=Real32Attribute(tab,"XOFFPNM2")
        end if
      end if
      if (hasAttribute(set,"YOFFPNM1")) then    
        yoff(2)=Real32Attribute(set,"YOFFPNM1")
      else 
        tab = table(set,"SRCLIST")
        if (hasAttribute(tab,"YOFFPNM1")) then
          yoff(2)=Real32Attribute(tab,"YOFFPNM1")
        end if
      end if
      if (hasAttribute(set,"YOFFPNM2")) then    
        yoff(3)=Real32Attribute(set,"XOFFPNM2")
      else 
        tab = table(set,"SRCLIST")
        if (hasAttribute(tab,"YOFFPNM2")) then
          yoff(3)=Real32Attribute(tab,"YOFFPNM2")
        end if
      end if
      call release(set)
    end if

    ecf=1.0
    necf=parameterCount("ecf")
    do i=1, necf
      ecf(i) = realParameter("ecf",i-1)
    end do

    ict_ech = 1
!    eimage(1) = c_image(1)
    dm_nrun = num_runs
    dm_larx = boxsize+1
    dm_lary = boxsize+1
    dm_prio = 1
    dm_snrm = likmin
    dm_lmap = maskflg
    dm_lloc = locflg
    dm_lexp = expoflg
    bscal = 1.

    call get_files("imagesets", imag, instrume, band, .true.)
    c_image = imag 
    eimage(1) = c_image(1)
    if (.not. locflg) then  
      call get_files("bkgimagesets", imag, instrume, band, .false.)
      c_bgima = imag
      bacmp(1) = c_bgima(1)
    end if
    if (expoflg) then
      call get_files("expimagesets", imag, instrume, band, .false.)     
      c_expima = imag
      expom(1) = c_expima(1) 
    end if
    if (maskflg) then
      call get_files("detmasksets", imag, instrume, band, .false.) 
      c_detima = imag
      mexbm = c_detima(1)
    end if
    
    if (necf/=nim) call warning("FileMismatch", &
      "Number of ECFs not equal number of images") 

    npi=parameterCount("pimin")
    pimin=2000.0
    do i=1,npi 
      pimin(i)=intParameter("pimin",i-1)
    end do
    if (npi/=nim) call warning("FileMismatch", &
      "Number of PI boundaries not equal number of images") 

    npi=parameterCount("pimax")
    pimax=4500.0
    do i=1,npi 
      pimax(i)=intParameter("pimax",i-1)
    end do
    if (npi/=nim) call warning("FileMismatch", &
      "Number of PI boundaries not equal number of images") 
    
    do i=1,nim
      energy(i)=(pimin(i)+pimax(i))/2.0
    end do    



  end subroutine dmpari

  !---------------------------------------------------------------------
  subroutine wr_stab(ll,nsrc,xms,yms,exms,eyms,scts,escts,bkg,exi_ml, &
                     box_fact,raw_cts,raw_bkg,eef,maskzero) 
  !---------------------------------------------------------------------

    integer, intent(in) :: nsrc 
!    real, dimension(max_sou), intent(in) :: xms, yms, exms, eyms 
    real, dimension(nsrc), intent(in) :: xms, yms, exms, eyms 

    real, dimension(:,:), intent(in) :: scts,escts,exi_ml,bkg, &
                                        raw_cts,raw_bkg,eef
    integer, dimension(:,:), intent(in) :: maskzero

    real, dimension(nsrc) :: raw_ctst,raw_bkgt,eeft
    real, dimension(:,:,:), allocatable ::raw_cts0,raw_bkg0,eef0,vig
    real, dimension(max_ins,nsrc) ::  raw_ctsi,raw_bkgi,eefi
    integer, dimension(nsrc), intent(in) :: box_fact
    integer, dimension(nsrc) :: box_size
    real, dimension(nsrc) :: xms_f, yms_f 
    real(kind=single), dimension(nsrc)  :: eradec 
    real, dimension(:,:), allocatable :: expos, flx, rate, eflx, erate

    real, dimension(max_ins,nsrc) :: sctsi, esctsi, flxi, eflxi, &
                                        ratei, eratei, bkgi, expoi, exi_mli  
    real, dimension(:,:,:), allocatable :: scts0, escts0, &
                                flx0, eflx0, rate0, erate0, expos0, &
                                bkg0, exi_ml0, ext_ml0
    integer, dimension(:,:,:), allocatable :: nzero0
    real, dimension(nsrc) :: sctst, esctst, flxt, eflxt, ratet, &
                                eratet, bkgt, expot, exi_mlt
    real(kind=single), dimension(max_ins,nsrc) :: hr1, hr2, hr3, &
                                                     hr1_err, hr2_err, hr3_err
    real(kind=single) :: hew, sexp, ootfac
    real(kind=double) :: ddec, dra, d0r, r0r, rr, dr, xr, yr, &
                         lr, br, cos_pa, sin_pa
    real(kind=double), dimension(nsrc) :: ra, dec, lii, bii
    integer :: num_instr0, num_image0(max_ima), nflux, bs
    integer :: nexp,ix0,iy0,ix,iy,i,i0,j,j0,k,l,ll,ipoint,instused,max_image
    integer ::  iimage,iinstr,ieband,nzero 
    integer, dimension(3) :: nf, ng
    integer, save :: id_src = 0
    logical, dimension(0:4) :: instflg

    !!! IT 2015/06 for input images in detector coordinates
    real, dimension(nsrc) :: xms_det, yms_det
    real(kind=double), dimension(nsrc) :: xtest,ytest
    type(DataSetT) :: inset
    character(len=80) :: inima
    integer(int32) :: ininstid
    real(double) :: intimestamp
    type(SpacecraftAttitudeType) :: inscAttitude, ininstrAttitude

! Determine rates, fluxes, sums, per band and total
    

    nf=0
    ng=0
    max_image=0
    do j0=1,num_instr
      j=expip(j0)
      nf(j)=nf(j)+num_image(j0)
      ng(j)=ng(j)+1
      max_image=max(num_image(j0),max_image)
    end do


    allocate (scts0(3,max_image,nsrc))
    allocate (escts0(3,max_image,nsrc))
    allocate (flx0(3,max_image,nsrc))
    allocate (eflx0(3,max_image,nsrc))
    allocate (rate0(3,max_image,nsrc))
    allocate (erate0(3,max_image,nsrc))
    allocate (expos0(3,max_image,nsrc))
    allocate (bkg0(3,max_image,nsrc))
    allocate (exi_ml0(3,max_image,nsrc))
    allocate (ext_ml0(3,max_image,nsrc))
    allocate (raw_cts0(3,max_image,nsrc))
    allocate (raw_bkg0(3,max_image,nsrc))
    allocate (eef0(3,max_image,nsrc))
    allocate (vig(3,max_image,nsrc))
    allocate (nzero0(3,max_image,nsrc))

    allocate (expos(nim,nsrc))
    allocate (flx(nim,nsrc))
    allocate (rate(nim,nsrc))
    allocate (eflx(nim,nsrc))
    allocate (erate(nim,nsrc))

    sctsi(1:num_instr,1:nsrc)=0.
    esctsi(1:num_instr,1:nsrc)=0.
    flxi(1:num_instr,1:nsrc)=0.
    eflxi(1:num_instr,1:nsrc)=0.
    ratei(1:num_instr,1:nsrc)=0.
    eratei(1:num_instr,1:nsrc)=0.
    raw_ctsi(1:num_instr,1:nsrc)=0.
    raw_bkgi(1:num_instr,1:nsrc)=0.
    eefi(1:num_instr,1:nsrc)=0.
    exi_mli=0. 
    exi_mlt=0.
    expoi(1:3,1:nsrc)=0.
    expot(1:nsrc)=0.
    bkgi(1:3,1:nsrc)=0.
    bkgt(1:nsrc)=0.
    sctst(1:nsrc)=0.
    esctst(1:nsrc)=0.
    raw_ctst(1:nsrc)=0.
    raw_bkgt(1:nsrc)=0.
    eeft(1:nsrc)=0.
    flxt(1:nsrc)=0.
    eflxt(1:nsrc)=0.
    ratet(1:nsrc)=0.
    eratet(1:nsrc)=0.
    scts0=0.
    escts0=0.
    raw_cts0=0.
    raw_bkg0=0.
    flx=0.
    eflx=0.
    rate=0.
    erate=0. 
    eef0=0.
    flx0=0.
    eflx0=0.
    rate0=0.
    erate0=0.
    exi_ml0=0.
    ext_ml0=0.
    nzero0=0
    bkg0=0.
    expos0=0.
    instflg(0)=.true.
    instflg(1:4)=.false.
    num_instr0=0
    hr1=-999.
    hr1_err=-999.
    hr2=-999.
    hr2_err=-999.
    hr3=-999.
    hr3_err=-999.

    !!! IT 2015/06 coordinate conversion detcoord -> radec via cal
    if (indetcoords) then
       inima = stringParameter("imagesets",0)
       inset = dataSet (inima, READ)
       ! in detector coordinates, only one instrument is allowed, so we can
       ! just read it from the first image (code copied from eboxdetect_mod)
       if (index(instrume(1),"PN") > 0 .or. &
            index(instrume(1),"pn") > 0) then 
          call CAL_setState1(instrument=EPN,AccuracyLevel=ACCURACY_MEDIUM)   
       elseif (index(instrume(1),"MOS1") > 0 .or. &
            index(instrume(1),"mos1") > 0) then 
          call CAL_setState1(instrument=EMOS1,AccuracyLevel=ACCURACY_MEDIUM) 
       else 
          call CAL_setState1(instrument=EMOS2,AccuracyLevel=ACCURACY_MEDIUM) 
       endif
       call getCalInfo(inset, ininstid, intimestamp, inscAttitude, ininstrAttitude)
       ! conversion
       xms_det(:)=(xms(:)+0.5-crpix1) * crdelt1
       yms_det(:)=(yms(:)+0.5-crpix2) * crdelt2
       call detToRaDec(dble(xms_det), dble(yms_det), inscAttitude, intimestamp, ra, dec)
       ra(:)=ra(:)*deg ! rad = pi/180.d0 ; deg = 1d0/rad
       dec(:)=dec(:)*deg
       ! eradec: use binning factor (crdelt1/2) and pixel size (detector units
       ! in arcsec: 0.05)
       eradec(:) = sqrt((exms(:)*crdelt1*0.05d0)**2+(eyms(:)*crdelt2*0.05d0)**2)
    else
       ! initialize xms_det and yms_det with zero (... dummy ...)
       xms_det(:)=0.0
       yms_det(:)=0.0
       ! calculate constants only once ...
       cos_pa = dcos(crota2 * rad)
       sin_pa = dsin(crota2 * rad)
    endif

    do i=1,nsrc
      box_size(i)=box_fact(i)*(boxsize*2+1)
      do j0=1,num_instr
        j=expip(j0)
        instflg(j)=.true.
        num_instr0=max(j,num_instr0)
        num_image0(j)=num_image(j0)
        do l=1,num_image(j0)
          ipoint=ip(j0,l)

          ix0 = nint(xms(i)+0.5)
          iy0 = nint(yms(i)+0.5)
            sexp=0.0
            nexp = 0
            do ix=ix0-1,ix0+1
              do iy=iy0-1,iy0+1
                if (exp_image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) > 0.0) then
                  sexp=sexp+exp_image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1))
                end if
                nexp=nexp+1
              end do
            end do
            expos(ipoint,i) = sexp/nexp

          OOTfac=1.0 
          if (index(instrume(ipoint),"PN")>0.or.index(instrume(ipoint),"pn")>0) then
            if (index(submode,"rime")>0.and.index(submode,"ull")>0.or. &
                index(submode,"RIME")>0.and.index(submode,"ULL")>0) OOTfac=0.9411
            if (index(submode,"Ext")>0.and.index(submode,"ull")>0.or. &
                index(submode,"ext")>0.and.index(submode,"ull")>0.or. &
                index(submode,"EXT")>0.and.index(submode,"ULL")>0) OOTfac=0.97815
          end if

          if (expos(ipoint,i)>0.0) then
            exi_mli(j,i)=exi_mli(j,i)+exi_ml(ipoint,i) 
            exi_mlt(i)=exi_mlt(i)+exi_ml(ipoint,i)

            flx(ipoint,i)  = scts(ipoint,i)/expos(ipoint,i)/ecf(ipoint)*1.e-11/OOTfac
            eflx(ipoint,i) = escts(ipoint,i)/expos(ipoint,i)/ecf(ipoint)*1.e-11/OOTfac
            rate(ipoint,i) = scts(ipoint,i)/expos(ipoint,i)/OOTfac
            erate(ipoint,i)= escts(ipoint,i)/expos(ipoint,i)/OOTfac
 
            scts0(j,l,i) = scts0(j,l,i)+scts(ipoint,i)
            escts0(j,l,i) = escts0(j,l,i)+escts(ipoint,i)**2
            raw_cts0(j,l,i) =  raw_cts0(j,l,i) +  raw_cts(ipoint,i)          
            raw_bkg0(j,l,i) =  raw_bkg0(j,l,i) +  raw_bkg(ipoint,i)
            eef0(j,l,i) = eef0(j,l,i) +  eef(ipoint,i)       

            flx0(j,l,i) = flx0(j,l,i)+flx(ipoint,i)*expos(ipoint,i) 
            eflx0(j,l,i) = eflx0(j,l,i)+(eflx(ipoint,i)*expos(ipoint,i))**2
            rate0(j,l,i) = rate0(j,l,i)+rate(ipoint,i)*expos(ipoint,i)
            erate0(j,l,i) = erate0(j,l,i)+(erate(ipoint,i)*expos(ipoint,i))**2
            exi_ml0(j,l,i) = exi_ml0(j,l,i)+exi_ml(ipoint,i)
            nzero0(j,l,i) = maskzero(ipoint,i)
            bkg0(j,l,i) = bkg0(j,l,i)+bkg(ipoint,i)
            expos0(j,l,i) = expos0(j,l,i)+expos(ipoint,i)

            sctsi(j,i)=sctsi(j,i)+scts(ipoint,i) 
            esctsi(j,i)=esctsi(j,i)+escts(ipoint,i)**2
            raw_ctsi(j,i) =  raw_ctsi(j,i) +  raw_cts(ipoint,i)          
            raw_bkgi(j,i) =  raw_bkgi(j,i) +  raw_bkg(ipoint,i)

            sctst(i)=sctst(i)+scts(ipoint,i)
            esctst(i)=esctst(i)+escts(ipoint,i)**2
            raw_ctst(i) =  raw_ctst(i) +  raw_cts(ipoint,i)          
            raw_bkgt(i) =  raw_bkgt(i) +  raw_bkg(ipoint,i)

            vig(j,l,i)=expos(ipoint,i)/exp_center(ipoint)
            bkgi(j,i)=bkgi(j,i)+bkg(ipoint,i)
            bkgt(i)=bkgt(i)+bkg(ipoint,i)
            expoi(j,i)=expoi(j,i)+expos(ipoint,i)
            expot(i)=expot(i)+expos(ipoint,i)

          end if 
        end do

        esctsi(j,i)=sqrt(esctsi(j,i))
      end do

!      do j0=1,num_instr
!        j=expip(j0)
       do j=1,3 
        do l=1,num_image(1)
          if (expos0(j,l,i)>0.) then
            escts0(j,l,i) = sqrt(escts0(j,l,i))
            rate0(j,l,i) = rate0(j,l,i)/expos0(j,l,i)
            erate0(j,l,i) = sqrt(erate0(j,l,i))/expos0(j,l,i)
            flx0(j,l,i) = flx0(j,l,i)/expos0(j,l,i)
            eflx0(j,l,i) = sqrt(eflx0(j,l,i))/expos0(j,l,i)            
          end if
        end do
      end do

!      do j0 = 1, num_instr
!        j=expip(j0)
      do j=1,3  
        if (instflg(j)) then
          do l=1,num_image(j0)
            ratei(j,i)=ratei(j,i)+rate0(j,l,i)
            eratei(j,i)=eratei(j,i)+erate0(j,l,i)**2
            flxi(j,i)=flxi(j,i)+flx0(j,l,i)
            eflxi(j,i)=eflxi(j,i)+eflx0(j,l,i)**2
          end do
        end if
        eratei(j,i)=sqrt(eratei(j,i))
        eflxi(j,i)=sqrt(eflxi(j,i))
      end do 

!      do j0=1,num_instr
!        j=expip(j0)
      do j=1,3
        if (instflg(j)) then
          ratet(i)=ratet(i)+ratei(j,i)
          eratet(i)=eratet(i)+eratei(j,i)**2
          flxt(i)=flxt(i)+flxi(j,i)/eflxi(j,i)**2
          eflxt(i)=eflxt(i)+1./eflxi(j,i)**2
        end if
      end do 

      esctst(i)=sqrt(esctst(i))
      eratet(i)=sqrt(eratet(i))
      flxt(i)=flxt(i)/eflxt(i)
      eflxt(i)=1./sqrt(eflxt(i))

!!!!! correct dof for images where mask was zero:
      nzero=0
      do iimage=1,nim
        nzero=nzero+maskzero(iimage,i) 
      end do
      exi_mlt(i)=gammq(real(nim-nzero),exi_mlt(i))
      nzero=0
      do j=1,3
        if (ng(j)>0) then
!!!!!!    correct dof for images where mask was zero:
          nzero=0
          do iinstr=1,max_ins
            do ieband=1,max_bands
              if (expip(iinstr) == j .and. ip(iinstr,ieband) > 0) then
                 nzero=nzero+maskzero(ip(iinstr,ieband),i)
              endif  
            end do
          end do

          exi_mli(j,i)=gammq(real(nf(j)-nzero),exi_mli(j,i))
          do l=1,num_image0(j)
            exi_ml0(j,l,i)=gammq(real(ng(j)-nzero0(j,l,i)),exi_ml0(j,l,i))
          end do
        end if 
      end do

      xms_f(i) = xms(i)+0.5
      yms_f(i) = yms(i)+0.5 

      ! standard coordinate conversion
      if (.not. indetcoords) then
         xr = (xms_f(i)-crpix1) * crdelt1 * rad
         yr = (yms_f(i)-crpix2) * crdelt2 * rad
         dra =  xr * cos_pa - yr * sin_pa
         ddec = xr * sin_pa + yr * cos_pa
         r0r = crval1 * rad
         d0r = crval2 * rad
         call antisp(dra,ddec,r0r,d0r,rr,dr)
         ra(i) =  rr * deg
         dec(i) = dr *deg

         eradec(i)  = sqrt((exms(i)*crdelt1)**2+(eyms(i)*crdelt2)**2)*3600.0
      else
         ! input for lii / bii:                        !!! IT 2015/06
         rr = ra(i)  * rad
         dr = dec(i) * rad
      endif


      call sla_EQGAL(rr,dr,lr,br)
      lii(i) = lr * deg
      bii(i) = br * deg
    end do

    do j0=1,num_instr
     j=expip(j0)
     do i=1,nsrc
      if (hrdef(1)>0.and.hrdef(2)>0.and. &
          max(hrdef(1),hrdef(2))<=num_image(j0)) then    
       hr1(j,i)=(max(0.,rate0(j,hrdef(2),i))-max(0.,rate0(j,hrdef(1),i)))/ &
                (max(0.,rate0(j,hrdef(2),i))+max(0.,rate0(j,hrdef(1),i)))
       hr1_err(j,i)=2.*sqrt((rate0(j,hrdef(1),i)*erate0(j,hrdef(2),i))**2+ &
                            (rate0(j,hrdef(2),i)*erate0(j,hrdef(1),i))**2)/&
                (max(0.,rate0(j,hrdef(2),i))+max(0.,rate0(j,hrdef(1),i)))**2
      end if
      if (hrdef(3)>0.and.hrdef(4)>0.and. &
          max(hrdef(3),hrdef(4))<=num_image(j0)) then    
       hr2(j,i)=(max(0.,rate0(j,hrdef(4),i))-max(0.,rate0(j,hrdef(3),i)))/ &
                (max(0.,rate0(j,hrdef(4),i))+max(0.,rate0(j,hrdef(3),i))) 
       hr2_err(j,i)=2.*sqrt((rate0(j,hrdef(3),i)*erate0(j,hrdef(4),i))**2+ &
                            (rate0(j,hrdef(4),i)*erate0(j,hrdef(3),i))**2)/&
                (max(0.,rate0(j,hrdef(4),i))+max(0.,rate0(j,hrdef(3),i)))**2
      end if
      if (hrdef(5)>0.and.hrdef(6)>0.and. &
          max(hrdef(5),hrdef(6))<=num_image(j0)) then    
       hr3(j,i)=(max(0.,rate0(j,hrdef(6),i))-max(0.,rate0(j,hrdef(5),i)))/ &
                (max(0.,rate0(j,hrdef(6),i))+max(0.,rate0(j,hrdef(5),i)))
       hr3_err(j,i)=2.*sqrt((rate0(j,hrdef(5),i)*erate0(j,hrdef(6),i))**2+ &
                            (rate0(j,hrdef(6),i)*erate0(j,hrdef(5),i))**2)/&
                (max(0.,rate0(j,hrdef(6),i))+max(0.,rate0(j,hrdef(5),i)))**2
      end if
     end do
    end do

    hew = 0. 
 
    ! Write out results
    instused = 0
    do i=1,4
      if (instflg(i)) instused=instused+1
    end do

    ! Instrument summary only required if more than one instrument
    if (instused > 1) then
      i0 = 0
    else
      i0 = 1
    end if

    do k = 1,nsrc

      if (exi_mlt(k) >= dm_snrm) then  

        id_src = id_src+1

        do i = i0,num_instr0

         if (instflg(i)) then 

          if (i == 0) then

          ! Write instrument summary
            call boxout(nsrc,1,ll,id_src,i,0, &
               sctst(k),esctst(k),raw_ctst(k),xms_f(k),yms_f(k),exms(k),eyms(k), &
               exi_mlt(k),bkgt(k),raw_bkgt(k),expot(k),flxt(k),eflxt(k), &
               ratet(k),eratet(k), &
               ra(k),dec(k),eradec(k),lii(k), bii(k),0.,0.,0.,0.,0.,0.,0., &
               box_size(k),eeft(k), &
               xms_det(k),yms_det(k))

          else
          ! Energy band summary only required if more than one band
            if (num_image0(i) > 1) then
              j0 = 0

            else
              j0 = 1
            end if

            do j = j0,num_image0(i)

  
              ! Write energy band summary
              if (j == 0) then
                call boxout(nsrc,2,ll,id_src,i,j, &
                   sctsi(i,k),esctsi(i,k),raw_ctsi(i,k),xms_f(k), &
                   yms_f(k),exms(k),eyms(k), &
                   exi_mli(i,k),bkgi(i,k),raw_bkgi(i,k),expoi(i,k),flxi(i,k), &
                   eflxi(i,k),ratei(i,k),eratei(i,k),ra(k),dec(k),eradec(k), &
                   lii(k),bii(k),hr1(i,k),hr1_err(i,k),hr2(i,k),hr2_err(i,k), &
                   hr3(i,k),hr3_err(i,k),0.,box_size(k),eefi(i,k), &
                   xms_det(k),yms_det(k))
              end if

              ! Write results per band and per instrument
              if (j /= 0) then
                ipoint=ip(i,j)
                call boxout(nsrc,3,ll,id_src,i,j, &
                   scts0(i,j,k),escts0(i,j,k),raw_cts0(i,j,k),xms_f(k),yms_f(k), &
                   exms(k),eyms(k),exi_ml0(i,j,k),bkg0(i,j,k),raw_bkg0(i,j,k), &
                   expos0(i,j,k),flx0(i,j,k),eflx0(i,j,k), &
                   rate0(i,j,k),erate0(i,j,k),ra(k),dec(k),eradec(k), &
                   lii(k),bii(k),0.,0.,0.,0.,0.,0.,vig(i,j,k),box_size(k), &
                   eef0(i,j,k), &
                   xms_det(k),yms_det(k))
              end if 
 
            end do      
          end if
         end if
        end do
      end if 
    end do

  end subroutine wr_stab

  !---------------------------------------------------------------------
  subroutine boxout (nsrc,k,ll,id_src_d2,id_inst,id_band,&
                    scts,scts_err,box_cts,x_ima,y_ima,x_ima_err,y_ima_err,    &
                    exi_ml,bg_map,bg_raw, &
                    exp_map,flux,flux_err,rate,rate_err,ra,dec,       &
                    radec_err,lii,bii,hr1,hr1_err,hr2,hr2_err,hr3,hr3_err, &
                    vig,box_size,eef,  xdet,ydet)
  !---------------------------------------------------------------------
        
    integer, intent(IN) :: k,id_src_d2,id_inst, &
                           id_band
    real(kind=single), intent(IN) :: scts,scts_err,x_ima,y_ima,        &
                        x_ima_err,y_ima_err,exi_ml, &
                        bg_map,exp_map,flux,flux_err,rate,rate_err,    &
                        radec_err,hr1,hr1_err,hr2,hr2_err,hr3,hr3_err, &
                        vig,   xdet,ydet    !!! IT 2015/06: detector coordinates
    real(kind=double), intent(IN) :: ra,dec,lii,bii
    integer, intent(IN) :: box_size,nsrc
    integer, intent(INOUT) :: ll
    real(kind=single),  intent(IN) :: box_cts,bg_raw,eef

    integer, parameter :: ncol = 33
    integer :: i, j, l1, nbuff

    type(DataSetT), save :: Set
    type(DataSetT) :: set0
    type(TableT), save :: tab
    type(ColumnT), dimension(35), save :: c 

    integer, dimension(:), allocatable, save :: ia0,ia1,ia2,ia3,ia4,ia5 
    real(kind=single), dimension(:), allocatable, save :: ra6,ra7,ra8, &
      ra9,ra10,ra11,ra12,ra13,ra14,ra15,ra16,ra17,ra18,ra19, &
      ra20,ra21,ra24,ra25,ra26,ra27,ra28,ra29,ra30,ra31,ra32, &
      ra33,ra34,ra35,ra36, ra37,ra38   !!! IT 2015/06: rp37,rp38 - detx,dety
    real(kind=double), dimension(:),allocatable,save :: da22,da23,da34,da35

    integer, dimension(:), pointer :: ip1, ip3, ip4 
    real(kind=single), dimension(:), pointer :: rp6, rp7, rp8, &
      rp9, rp10, rp11, rp14, rp16, rp17, rp18, rp19, &
      rp20, rp21, rp24, rp25, rp26, rp27, rp28, rp29, rp30, rp31, rp32, &
      rp33, rp34, rp35, rp36, rp37, rp38
    real(kind=double), dimension(:), pointer :: dp22, dp23, dp34, dp35
    real(kind=double), dimension(max_ins) :: ra_cen, dec_cen
    real(kind=double) :: xr, yr, rr, dr, dra, ddec, r0r, d0r, cos_pa, sin_pa
    character(len=8) :: rakey, dekey
    character(len=2), dimension(3) :: ins

    logical, save :: first=.true.


    if (first) then
      first =.false.
!      nbuff=max_sou*(nim+num_instr+1)
       nbuff=nsrc*(nim+num_instr+1)
      allocate(ia0(0:nbuff))
      allocate(ia1(0:nbuff))
      allocate(ia2(0:nbuff))
      allocate(ia3(0:nbuff))
      allocate(ia4(0:nbuff))
      allocate(ia5(0:nbuff))
      allocate(ra6(0:nbuff))
      allocate(ra7(0:nbuff))
      allocate(ra8(0:nbuff))
      allocate(ra9(0:nbuff))
      allocate(ra10(0:nbuff))
      allocate(ra11(0:nbuff))
      allocate(ra12(0:nbuff))
      allocate(ra13(0:nbuff))
      allocate(ra14(0:nbuff))
      allocate(ra15(0:nbuff))
      allocate(ra16(0:nbuff))
      allocate(ra17(0:nbuff))
      allocate(ra18(0:nbuff))
      allocate(ra19(0:nbuff))
      allocate(ra20(0:nbuff))
      allocate(ra21(0:nbuff))
      allocate(ra24(0:nbuff))
      allocate(ra25(0:nbuff))
      allocate(ra26(0:nbuff))
      allocate(ra27(0:nbuff))
      allocate(ra28(0:nbuff))
      allocate(ra29(0:nbuff))
      allocate(ra30(0:nbuff))
      allocate(ra31(0:nbuff))
      allocate(ra32(0:nbuff))
      allocate(ra33(0:nbuff))
      allocate(ra34(0:nbuff))
      allocate(ra35(0:nbuff))
      allocate(ra36(0:nbuff))
      allocate(ra37(0:nbuff))
      allocate(ra38(0:nbuff))
      allocate(da22(0:nbuff))
      allocate(da23(0:nbuff))
      allocate(da34(0:nbuff))
      allocate(da35(0:nbuff))
    end if

    if (k < 0) then

!      ll=-ll
      l1=0     
      set = dataSet (stringParameter("boxlistset"), CREATE)
      set0 = dataSet (c_image(1), READ) 
       
      if (dm_lloc) then
        call cp_head (set0, set, "EPIC EXPOSURE BOX-LOCAL SOURCE LIST", 3)
        call setAttribute(set,"BOXMODE", "LOCAL", "EBOXDETECT mode")
      else
        call cp_head (set0, set, "EPIC EXPOSURE BOX-MAP SOURCE LIST", 3)      
        call setAttribute(set,"BOXMODE", "MAP", "EBOXDETECT mode")
      end if    
      call release(set0)
      call setAttribute(set, "HDUCLASS", "OGIP    ", &
                              "Format conforms to OGIP/GSFC conventions")
      call setAttribute(set, "HDUCLAS1", "SRCLIST   ", &
                              "File contains a source list")

      ins(1) = "PN"
      ins(2) = "M1"
      ins(3) = "M2"
      do j=1,num_instr
        cos_pa = dcos(crota2 * pi/180d0)
        sin_pa = dsin(crota2 * pi/180d0)
        xr = (xcenters(j)-crpix1) * crdelt1 * pi/180d0
        yr = (ycenters(j)-crpix2) * crdelt2 * pi/180d0
        dra =  xr * cos_pa - yr * sin_pa
        ddec = xr * sin_pa + yr * cos_pa
        r0r = crval1 * pi/180d0
        d0r = crval2 * pi/180d0
        call antisp(dra,ddec,r0r,d0r,rr,dr)
        ra_cen(j) = rr * 180d0/pi
        dec_cen(j) = dr *180d0/pi

        write (rakey,'("OARA",a,i2.2)')ins(expip(j)),j
        write (dekey,'("OADE",a,i2.2)')ins(expip(j)),j
        call setAttribute(set, rakey, ra_cen(j), "Optical axis: RA [deg]") 
        call setAttribute(set, dekey, dec_cen(j), "Optical axis: DEC [deg]")
      end do

      tab = addTable (set, "SRCLIST", ll)

!      l1 = l1+1
!      c(l1) = addColumn (tab, "SRC_NUM",  Integer32, " ")
!      ip0 => int32Data(c(l1))
!      ip0 = ia0(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "BOX_ID_SRC",  Integer32, " ")
      ip1 => int32Data(c(l1))
      ip1 = ia1(0:ll-1)

!      l1 = l1+1
!      c(l1) = addColumn (tab, "ID_SRC_D1",  Integer32, " ")
!      ip2 => int32Data(c(l1))
!      ip2 = ia2(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "ID_INST",    Integer32, " ")
      ip3 => int32Data(c(l1))
      ip3 = ia3(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "ID_BAND",    Integer32, " ")
      ip4 => int32Data(c(l1))
      ip4 = ia4(0:ll-1)

!      l1 = l1+1
!      c(l1) = addColumn (tab, "ID_CLUSTER", Integer32, " ")
!      ip5 => int32Data(c(l1))
!      ip5 = ia5(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "SCTS",       Real32, "counts")
      rp6 => real32Data(c(l1))
      rp6 = ra6(0:ll-1)
      do i=0,ll-1        
        if (ra17(i)<=0.) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "SCTS_ERR",   Real32, "counts")
      rp7 => real32Data(c(l1))
      rp7 = ra7(0:ll-1)
      do i=0,ll-1        
        if (ra17(i)<=0.) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "BOX_CTS",   Real32, "counts")
      rp33 => real32Data(c(l1))
      rp33 = ra33(0:ll-1)
      do i=0,ll-1        
        if (ra17(i)<=0.) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "X_IMA",      Real32, "image pixels")
      rp8 => real32Data(c(l1))
      rp8 = ra8(0:ll-1) 

      l1 = l1+1
      c(l1) = addColumn (tab, "X_IMA_ERR",  Real32, "image pixels")
      rp9 => real32Data(c(l1))
      rp9 = ra9(0:ll-1)
   
      l1 = l1+1
      c(l1) = addColumn (tab, "Y_IMA",      Real32, "image pixels")
      rp10 => real32Data(c(l1))
      rp10 = ra10(0:ll-1) 

      l1 = l1+1
      c(l1) = addColumn (tab, "Y_IMA_ERR",  Real32, "image pixels")
      rp11 => real32Data(c(l1))
      rp11 = ra11(0:ll-1) 

      l1 = l1+1
      c(l1) = addColumn (tab, "LIKE",     Real32, " ")
      rp14 => real32Data(c(l1))
      rp14 = ra14(0:ll-1)
      do i=0,ll-1        
        if (ra17(i)<=0.) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "BG_MAP",     Real32, "counts/pixel")
      rp16 => real32Data(c(l1))
      rp16 = ra16(0:ll-1)
      do i=0,ll-1        
        if (ra17(i)<=0.) call settoNull(c(l1),i)
      end do
       
     
      l1 = l1+1
      c(l1) = addColumn (tab, "BG_RAW",     Real32, "counts/pixel")
      rp34 => real32Data(c(l1))
      rp34 = ra34(0:ll-1)      
      do i=0,ll-1        
        if (ra17(i)<=0.) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "EXP_MAP",    Real32, "seconds")
      rp17 => real32Data(c(l1))
      rp17 = ra17(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)==0.or.ra17(i)<=0.) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "FLUX",       Real32, "cgs")
      rp18 => real32Data(c(l1))
      rp18 = ra18(0:ll-1)
      do i=0,ll-1        
        if (ra17(i)<=0.) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "FLUX_ERR",   Real32, "cgs")
      rp19 => real32Data(c(l1))
      rp19 = ra19(0:ll-1) 
      do i=0,ll-1        
        if (ra17(i)<=0.) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "RATE",       Real32, "counts/pixel")
      rp20 => real32Data(c(l1))
      rp20 = ra20(0:ll-1)
      do i=0,ll-1        
        if (ra17(i)<=0.) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "RATE_ERR",   Real32, "counts/pixel")
      rp21 => real32Data(c(l1))
      rp21 = ra21(0:ll-1) 
      do i=0,ll-1        
        if (ra17(i)<=0.) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "RA",         Real64, "deg")
      dp22 => real64Data(c(l1))
      dp22 = da22(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "DEC",        Real64, "deg")
      dp23 => real64Data(c(l1))
      dp23 = da23(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "RADEC_ERR",  Real32, "arcsec")
      rp24 => real32Data(c(l1))
      rp24 = ra24(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "LII",        Real64, "deg")
      dp34 => real64Data(c(l1))
      dp34 = da34(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "BII",        Real64, "deg")
      dp35 => real64Data(c(l1))
      dp35 = da35(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "HR1",        Real32, " ")
      rp25 => real32Data(c(l1))
      rp25 = ra25(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)/=0.or.ra25(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "HR1_ERR",    Real32, " ")
      rp26 => real32Data(c(l1))
      rp26 = ra26(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)/=0 .or. ra26(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "HR2",        Real32, " ")
      rp27 => real32Data(c(l1))
      rp27 = ra27(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)/=0 .or. ra27(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "HR2_ERR",    Real32, " ")
      rp28 => real32Data(c(l1))
      rp28 = ra28(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)/=0 .or. ra28(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "HR3",        Real32, " ")
      rp29 => real32Data(c(l1))
      rp29 = ra29(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)/=0 .or. ra29(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "HR3_ERR",    Real32, " ")
      rp30 => real32Data(c(l1))
      rp30 = ra30(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)/=0 .or. ra30(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "VIGNETTING",     Real32, " ")
      rp36 => real32Data(c(l1))
      rp36 = ra36(0:ll-1)
      do i=0,ll-1        
        if (ia4(i)==0.or.ia3(i)==0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "BOX_SIZE", Real32, "pixels")
      rp31 => real32Data(c(l1))
      rp31 = ra31(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "EEF", Real32, "")
      rp35 => real32Data(c(l1))
      rp35 = ra35(0:ll-1)
      do i=0,ll-1        
        if (ia4(i) == 0) call settoNull(c(l1),i)
      end do

      call neighbour(ra8,ra10,ra32,ll)

      l1 = l1+1
      c(l1) = addColumn (tab, "DIST_NN", Real32, "arcsec")
      rp32 => real32Data(c(l1))
      rp32 = ra32(0:ll-1)
      
      !!! IT 2015/06
      !!! DETX, DETY columns added as last columns of the output source list
      !!! if and only if input images are in detector coordinates. All other
      !!! columns - and column numbers - remain unchanged for consistency.
      if (indetcoords) then
         l1 = l1+1
         c(l1) = addColumn (tab, "DETX",      Real32, "detector pixels")
         rp37 => real32Data(c(l1))
         rp37 = ra37(0:ll-1) 

         l1 = l1+1
         c(l1) = addColumn (tab, "DETY",      Real32, "detector pixels")
         rp38 => real32Data(c(l1))
         rp38 = ra38(0:ll-1) 
      endif

      do i=1,l1
        call release(c(l1))
      end do
      call release (tab)
      call release (set)

    else    

      ia1(ll) = id_src_d2
      ia3(ll) = id_inst
      ia4(ll) = id_band
      ra6(ll) = scts
      ra7(ll) = scts_err
      ra8(ll) = x_ima
      ra9(ll) = x_ima_err
      ra10(ll) = y_ima 
      ra11(ll) = y_ima_err
      ra14(ll) = exi_ml
      ra16(ll) = bg_map
      ra17(ll) = exp_map
      ra18(ll) = flux
      ra19(ll) = flux_err
      ra20(ll) = rate
      ra21(ll) = rate_err
      da22(ll) = ra
      da23(ll) = dec
      ra24(ll) = radec_err
      da34(ll) = lii
      da35(ll) = bii
      ra25(ll) = hr1
      ra26(ll) = hr1_err
      ra27(ll) = hr2
      ra28(ll) = hr2_err 
      ra29(ll) = hr3
      ra30(ll) = hr3_err 
      ra31(ll) = box_size
      ra33(ll) = box_cts
      ra34(ll) = bg_raw
      ra35(ll) = eef
      ra36(ll) = vig
      if (indetcoords) then
         ra37(ll) = xdet
         ra38(ll) = ydet
      endif
      ll = ll+1 
      
    end if

  end subroutine boxout

!-------------------------------------------------------------------------
  subroutine neighbour (xms,yms,dist,nsrc)
!-------------------------------------------------------------------------

  integer, intent(in) :: nsrc 
!  real(kind=single), dimension(0:max_sou), intent(in) :: xms, yms
  real(kind=single), dimension(0:nsrc), intent(in) :: xms, yms
!  real(kind=single), dimension(0:max_sou), intent(out) :: dist
  real(kind=single), dimension(0:nsrc), intent(out) :: dist
  real(kind=single) d2,d2save
  integer :: i,j
 

 
  do i=0,nsrc-1
    d2save=1e10
    do j=0,nsrc-1
      d2=(xms(i)-xms(j))**2+(yms(i)-yms(j))**2
      if (d2 > 0.1 .and. d2 < d2save) then
        d2save=d2
      end if  
    end do
    dist(i)=sqrt(d2save)*abs(crdelt1)*3600.
  end do  

  end subroutine neighbour


  !---------------------------------------------------------------------
  subroutine rd_data
  !---------------------------------------------------------------------
 
    integer           i,j,ix,iy,ipoint,maxdim
    parameter         (maxdim=2)
    integer           npix0(2)  
    integer           dattyp,filtyp,iomode
    real              exposure
    double precision  start(maxdim),step(maxdim)
    logical           first
    character*1024      fname
 
    dattyp   = 10
    filtyp   = 1
    iomode   = 0


    nim = ip(num_instr,num_image(num_instr))
    first = .true.

    do i = 1,num_instr
      do j = 1,num_image(i)
        ipoint=ip(i,j)  

        fname = c_image(ipoint)
        call co_image(fname,start,step,npix,exposure,xcenters(ipoint),ycenters(ipoint), &
                      ra_pnt,dec_pnt,pa_pnt, &
                      crpix1,crpix2,crval1,crval2,crdelt1,crdelt2,crota2, &
                      xoff,yoff,expip(i),ipoint,nim)
       
         if (i == 1) then 
           npix0(1)=npix(1)
           npix0(2)=npix(2) 
         else
           if (npix0(1) /= npix(1) .or. npix0(2) /= npix(2)) then
             call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")   
           endif 
         endif

        if (.not.dm_lloc) then
          fname = c_bgima(ipoint)
          call co_bgima(fname,npix,xoff,yoff,expip(i),ipoint,nim)
        end if

        if (dm_lexp) then
          fname = c_expima(ipoint)
          call co_expima(fname,npix,xoff,yoff,expip(i),ipoint, &
                 xcenters(ipoint),ycenters(ipoint),exp_center(ipoint),nim) 
          if (npix0(1) /= npix(1) .or. npix0(2) /= npix(2)) then
            call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")   
          endif 

        else
          if (first) then
            allocate (exp_image(nim,imbuff,imbuff))
            first = .false.
          end if 
          do iy=1,npix(2)
            do ix=1,npix(1)
              exp_image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) = exposure
            end do
          end do
        end if
  
      end do

      if (dm_lmap) then
        ipoint=ip(i,1)
        fname = c_detima(i)
        call co_mask (fname, npix, xoff, yoff, expip(i),ipoint)
        if (npix0(1) /= npix(1) .or. npix0(2) /= npix(2)) then
          call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")   
        endif 
       
        do j = 2,num_image(i)
          do iy = 1,npix(2)
            do ix = 1,npix(1)
              mask_image(ip(i,j),ixb(ix,ip(i,j),1),iyb(iy,ip(i,j),1))= &
                mask_image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1))
            end do
          end do
        end do   
      end if 

    end do

    !!! IT 2015/06 for images in detector coordinates
    if (indetcoords) call warning("NoPSFcorrection", &
"Input images are in detector coordinates. Counts and fluxes in the output source list will not be corrected for PSF losses.")

   end subroutine rd_data

  !--------------------------------------------------------------------
  integer function lench(charr)
  !--------------------------------------------------------------------

! return position of last nonblank character

! author : zim               date: 07-mar-1983
! update : paul              date: 09-oct-1985 15:00
! update : paul              date: 1-jul-1987 16:41  
! update : exsas              date: 06-feb-1991 18:11

! lench returns length of nonblank content of character variable -
! blank= null,space,tab.

!  charr              c  i    character string to be checked
!  lench              i    o  length of nonblank content of string

    character*(*) charr
    character*1   nul ,tab
    integer       l
    nul = char(0)
    tab = char(9)

    l    = len(charr)
    lench= 0
    do while (lench.eq.0 .and. l.ge.1)
      if (charr(l:l).ne.' ' .and. charr(l:l).ne.tab .and. &
          charr(l:l).ne.nul) lench=l
        l= l-1
    enddo

    end function lench

end module ebox_io

