module eboxdata

  use dal
! use edetect_io

 implicit none     
  
!----------------------------------------------------------------------
! MAX_SOU              max number of sources
   integer, parameter :: max_sou = 500000
!----------------------------------------------------------------------
! cgenl: constants
!----------------------------------------------------------------------
!  outpt                      internal file for output handling
!  pi                         constant
!  rad                        for conversion of degrees into radians
!  deg                        for conversion of radians to degree
!  valmax                     maximum value for vax fortran
!  valmin                     minimum value for vax fortran
!  valeps                     precision for real*4 operations on vax 

      real(kind=double), parameter :: pi = 3.14159265358979323846d0, &
                         rad = pi/180d0, &
                         deg = 1d0/rad, &
                         valmax = 1.7e+38, &
                         valmin = 0.3e-37, &
                         valeps = 1.e-7

      character*256 outpt 
      character*132 endec


!----------------------------------------------------------------------
!  cdmio:  i/o files and parameters
!----------------------------------------------------------------------
!  dmexpd                     experient, detector
!  dmobsm                     pointing/survey
!  bacmp                      background file
!  image                      input file
!  bscal
!  dm_larx
!  dm_lary
!  dm_lmap                    flag to use the exposure map
!  dm_nrun
!  dm_nsrc                    output number of examined sources
!  dm_prio                    priority criterion for matching
!  dm_snrm                    minimum signal to noise ratio
!  idm_ech                    actual number of energy array files
!  idm_run
!  ict_bac                    
!  ict_img                    
!  ict_ech                    
!  ict_msb
!  ict_msl
!  inr_bac                    expected number of record of file bacmp
!  inr_img                    expected number of record of file image
!  inr_mxb                    expected number of record of file mexbm
!  ipa_ech                    maximum number of energy array files
!  ipa_run
!  lu_bac                     logical unit for file bacmp
!  lu_img                     logical unit for file image
!  lu_mxb                     logical unit for file mexbm
!  lu_msb                     logical unit for file msbld
!  lu_msl                     logical unit for file mslst
!  mexbm                      exposure map
!  msbld                      picture arrays
!  mslst                      output source list
!  nr_bac                     number of record of file bacmp
!  nr_img                     number of record of file image
!  nr_mxb                     number of record of file mexbm
!  nr_msb                     number of record of file msbld
!  nr_msl                     number of record of file mslst
!  ppois

      integer       ipa_ech, ipa_run
      parameter     (ipa_ech=18)
      parameter     (ipa_run=6)
      character*80  bacmp(ipa_ech), eimage(ipa_ech), ppois, mexbm,    &
                    expom(ipa_ech), mslst(ipa_ech), dmexpd, dmobsm
      real          bscal,dm_snrm ! ,ecf(ipa_ech)
      integer       ict_bac,ict_ech,ict_img,ict_msb,ict_msl,idm_ech, &
                    idm_run,inr_bac,inr_img(ipa_ech),inr_mxb,lu_bac, &
                    lu_img(ipa_ech),lu_msb(ipa_ech),lu_mxb,          &
                    lu_msl(ipa_ech),nr_bac,nr_img(ipa_ech),          &
                    nr_msb(ipa_ech),nr_msl(ipa_ech),nr_mxb,          &
                    dm_larx(ipa_run),dm_lary(ipa_run),dm_nrun,       &
                    dm_nsrc(ipa_ech),dm_prio(ipa_ech),boxsize,       &
                    hrdef(6) 
      logical       dm_lmap,dm_lloc, dm_lexp
      !!! IT 2015/03 new parameter obsmode: pointing or slew-mode data
      character(len=10) obsmode
      !!! IT 2015/06 new internal switch for images in detector coordinates
      logical       indetcoords

!??   idm_ech = ipa_ech
!??   idm_run = ipa_run

!----------------------------------------------------------------------
!  clslst: parameters for file lslst
!----------------------------------------------------------------------
!  abm_lsl                    distance to the next source in the field
!  abs_lsl                    distance to the next rib or edge
!  bkg_lsl                    background
!  clu_lsl                    cluster flag, represents the cluster size
!  con_lsl                    continuation flag (0 no continuat. line)
!  dat_lsl                    date of the file
!  dum_lsl                    dummy 
!  eof_lsl                    end of file flag
!  ext_lsl                    extent flag
!  fil_lsl                    list file name
!  i_lsl                      index of the source
!  ifl_lsl                    unused, always 1
!  snr_lsl                    signal to noise ratio
!  iwi_lsl                    size of detection window
!  mrk_lsl                    total number of sources in file
!  pri_lsl                    priority of the list
!  src_lsl                    source counts
!  tim_lsl                    day of the file
!  xms_lsl                    source x position
!  yms_lsl                    source y position

      character*80  fil_lsl
      character*12  dat_lsl, tim_lsl
      real          abs_lsl, abm_lsl, bkg_lsl, clu_lsl, dum_lsl, &
                    ext_lsl, src_lsl, xms_lsl, yms_lsl, snr_lsl
      integer       con_lsl,i_lsl,ifl_lsl,iwi_lsl,mrk_lsl, &
                    pri_lsl
      logical       eof_lsl

      integer       ipa_x, ipa_y, ipa_xy
!     parameter     (ipa_x = 640,  ipa_y = 640)
!      parameter     (ipa_xy = ipa_x*ipa_y)
      integer       idm_x, idm_y
!     parameter     (idm_x = ipa_x, idm_y = ipa_y)
      real, dimension(:,:), allocatable :: kwork,kfeld,ework,efeld,lfeld
      integer, dimension(:,:), allocatable :: ifeld, iwork 
      logical, dimension(:,:), allocatable :: mfeld, mwork

!----------------------------------------------------------------------
! csoio: i/o files
!----------------------------------------------------------------------
!  sopar                      name of instrument parameters for all
!  impar                      name of image parameter file
!  dlpar                      name of instrument parameters for exdl
!  dbpar                      name of instrument parameters for exdb
!  dmpar                      name of instrument parameters for exdm
!  smpar                      name of instrument parameters for exsm
!  sppar                      name of instrument parameters for exsp
!  ulpar                      name of parameters            for exul

      character*80      sopar,impar,dlpar,dbpar,dmpar,smpar,sppar,ulpar

!----------------------------------------------------------------------
! c_mslst:  parameters for file mslst
!----------------------------------------------------------------------
!  abs_msl                    distance to the next rib or edge
!  abm_msl                    distance to the next source in the field
!  bkg_msl                    background
!  clu_msl                    cluster flag, represents the cluster size
!  con_msl                    number of continuation line
!  dat_msl                    date of file's creation 
!  dum_msl                    dummy 
!  eof_msl                    end of file flag
!  ext_msl                    extent flag
!  fil_msl                    input file name
!  i_msl                      source index
!  ifl_msl                    always 0
!  snr_msl                    source signal to noise ratio
!  iwi_msl                    size of detection window
!  mrk_msl                    total number of sources
!  pri_msl                    priority of the list
!  src_msl                    source count
!  tim_msl                    time of the file's creation
!  xms_msl                    source x position
!  yms_msl                    source y position

      character*80  fil_msl
      character*12  dat_msl, tim_msl
      real          abs_msl, abm_msl, bkg_msl, clu_msl, dum_msl, &
                    ext_msl, src_msl, xms_msl, yms_msl, snr_msl
      integer       con_msl, i_msl, ifl_msl, iwi_msl, &
                    mrk_msl, pri_msl
      logical       eof_msl

end module eboxdata























































